import argparse
import sys

from sqlalchemy import create_engine, func
from sqlalchemy.orm import sessionmaker, aliased, joinedload
from sqlalchemy.sql.expression import true, false

from irco import graphs, models
from irco.logging import sentry


def write(fh, graph, format=None):
    from networkx.readwrite import gexf
    gexf.write_gexf(graph, fh)


def main():
    graph_choices = graphs.get_graphs()
    argparser = argparse.ArgumentParser('irco-graph')
    argparser.add_argument('-v', '--verbose', action='store_true')
    argparser.add_argument('-y', '--years')
    argparser.add_argument('-c', '--ca-country', action='append',
                           dest='ca_countries')
    argparser.add_argument('-t', '--type', action='append',
                           dest='types')
    argparser.add_argument('graph_type', choices=graph_choices)
    argparser.add_argument('database')
    argparser.add_argument('output', default='-', nargs='?')

    args = argparser.parse_args()

    sentry.context.merge({
        'tags': {
            'command': 'irco-graph',
            'graph_type': args.graph_type,
        },
        'extra': {'parsed_arguments': args.__dict__}
    })

    graph_factory = graphs.get_graph(args.graph_type)

    engine = create_engine(args.database, echo=args.verbose)
    Session = sessionmaker(bind=engine)
    session = Session()

    publications = (session
                    .query(models.Publication)
                    .options(
                        joinedload('affiliations').joinedload('author'),
                        joinedload('affiliations').joinedload('institution'),
                    ))

    if args.years:
        criteria = false()
        for y in args.years.split(','):
            if '-' in y:
                start, end = y.split('-')
                if start:
                    c = models.Publication.year >= int(start)
                else:
                    c = true()

                if end:
                    c = c & (models.Publication.year <= int(end))
            else:
                c = models.Publication.year == int(y)
            criteria = criteria | c

        publications = publications.filter(criteria)

    if args.ca_countries:
        c_author = aliased(models.AffiliatedAuthor)
        c_institution = aliased(models.Institution)

        criteria = false()

        for country in args.ca_countries:
            criteria = criteria | (
                func.lower(c_institution.country) == func.lower(country)
            )

        publications = (publications
                        .join(c_author)
                        .join(c_institution)
                        .filter(c_author.is_corresponding == True)
                        .filter(criteria))

    if args.types:
        criteria = false()

        for type in args.types:
            criteria = criteria | (models.Publication.type == type)

        publications = publications.filter(criteria)

    graph = graph_factory.create(session, publications)

    if args.output == '-':
        write(sys.stdout, graph)
    else:
        with open(args.output, 'wb') as fh:
            write(fh, graph)



import argparse
import sys
from tablib import formats

from irco import utils


def get_format(args, choices):
    format = args.format
    if not format:
        # Guess it from file extension
        try:
            format = args.output.rsplit('.', 1)[1]
        except IndexError:
            if args.output != '-':
                print 'Could not guess output format from filename.'
            else:
                print 'Cannot guess output format when writing to stdout.'
            print 'Please use the -f/--format option to specify it.'
            sys.exit(1)
        else:
            if format not in choices:
                print ('Could not guess output format from filename ({}).'
                       .format(format))
                print 'Please use the -f/--format option to specify it.'
                sys.exit(1)
    return format


def write(fh, dataset, format):
    fh.write(getattr(dataset, format))


def main():
    format_choices = [fmt.title for fmt in formats.available]
    argparser = argparse.ArgumentParser('irco-convert')
    argparser.add_argument('-f', '--format', choices=format_choices)
    argparser.add_argument('-r', '--records')
    argparser.add_argument('source', nargs='+')
    argparser.add_argument('output')

    args = argparser.parse_args()

    if args.records:
        records = [int(r.strip()) for r in args.records.split(',')]
    else:
        records = None
    format = get_format(args, format_choices)
    dataset = utils.get_dataset(args.source, records)

    if args.output == '-':
        write(sys.stdout, dataset, format)
    else:
        with open(args.output, 'wb') as fh:
            write(fh, dataset, format)



import argparse

from irco.explorer import app, database, filters
#from irco.logging import sentry


def main():
    argparser = argparse.ArgumentParser('irco-explorer')
    argparser.add_argument('-p', '--port', default=8000, type=int)
    argparser.add_argument('-i', '--interface', default='127.0.0.1')
    argparser.add_argument('-d', '--debug', action='store_true')
    argparser.add_argument('database')

    args = argparser.parse_args()

    #sentry.context.merge({
    #    'tags': {'command': 'irco-explorer'},
    #    'extra': {'parsed_arguments': args.__dict__}
    #})

    app.config['DATABASE'] = args.database
    database.init_app(app)
    filters.init_app(app)

    app.run(
        host=args.interface,
        port=args.port,
        debug=args.debug,
    )



import pytest
from collections import namedtuple
from irco.authors import Author, NamePart, levenshtein


AuthorTestSet = namedtuple('AuthorTestSet',
                           ['corresponding', 'expected', 'candidates'])

corresponding_authors = [
    (u'Ei-Shazly, M', 0, (u'EI-Shazly, M.', u'Eissa, B.')),
    (u'Pan, WH', 0, (u'Weihua, Pan',
                     u'Liao, Wanqing',
                     u'Khayhan, Kantarawee',
                     u'Hagen, Ferry',
                     u'Boekhout, Teun',
                     u'Khayhan, Kantarawee',
                     u'Hagen, Ferry',
                     u'Boekhout, Teun',
                     u'Khayhan, Kantarawee',
                     u'Wahyuningsih, Retno',
                     u'Sjam, Ridhawati',
                     u'Wahyuningsih, Retno',
                     u'Chakrabarti, Arunaloke',
                     u'Chowdhary, Anuradha',
                     u'Ikeda, Reiko',
                     u'Taj-Aldeen, Saad J.',
                     u'Khan, Ziauddin',
                     u'Imran, Darma',
                     u'Imran, Darma',
                     u'Sriburee, Pojana',
                     u'Chaicumpar, Kunyaluk',
                     u'Ingviya, Natnicha',
                     u'Mouton, Johan W.',
                     u'Curfs-Breuker, Ilse',
                     u'Meis, Jacques F.',
                     u'Klaassen, Corne H. W.',
                     u'Mouton, Johan W.',
                     u'Meis, Jacques F.')),
    (u'Abu-Shady, ASI', 2, (u'Al-Mudhaf, Humood F.',
                            u'Al-Hayan, Mohammad N.',
                            u'Abu-Shady, Abdel-Sattar I.',
                            u'Selim, Mustafa I.')),
    (u'Al-Tahan, ARM', 0, (u'Al-Tahan, Abdel-Rahman M.',
                           u'Al-Jumah, Mohammed A.',
                           u'Bohlega, Saeed M.',
                           u'Al-Shammari, Suhail N.',
                           u'Al-Sharoqi, Isa A.',
                           u'Dahdaleh, Maurice P.',
                           u'Hosny, Hassan M.',
                           u'Yamout, Bassem I.')),
    (u'Elassar, AZA', 0, (u'Elassar, Abdel-Zaher A.',
                          u'Al-Fulaij, Othman A.',
                          u'El-Sayed, Ahmed E. M.',
                          u'Elassar, Abdel-Zaher A.')),
    (u'Muller, HP', 0, (u'Mueller, Hans-Peter',
                        u'Barrieshi-Nusair, Kefah M.')),
    (u'Chen, HH', 0, (u'Chen, Hsiao-Hwa',
                      u'Xiao, Yang',
                      u'Chen, Hui',
                      u'Du, Xiaojiang',
                      u'Guizani, Mohsen'
                      )),
    (u'Mosaad, MES', 0, (u'Mosaad, M. El-Sayed',
                         u'Al-Hajeri, M.',
                         u'Al-Ajmi, R.',
                         u'Koliub, Abo. M.')),
    (u'Heo, MS', 4, (u'Harikrishnan, Ramasamy',
                     u'Moon, Young-Gun',
                     u'Kim, Man-Chul',
                     u'Kim, Ju-Sang',
                     u'Heo, Moon-Soo',
                     u'Jin, Chang-Nam',
                     u'Balasundaram, Chellam',
                     u'Azad, I. S.')),
    (u'Khedr, MEM', 0, (u'Khedr, M. -E. M.',
                        u'Chamkha, A. J.',
                        u'Bayomi, M.')),
]


@pytest.fixture(params=corresponding_authors)
def corresponding_author(request):
    corresponding, expected, candidates = request.param
    candidates = [Author(c) for c in candidates]
    expected = candidates[expected]
    return AuthorTestSet(Author(corresponding), expected, candidates)


def test_authors(corresponding_author):
    corresponding, expected, candidates = corresponding_author
    match = corresponding.find_best_match(candidates)
    assert match
    assert match == expected


def test_namepart_isabbrof():
    assert NamePart('WH').isabbrof(NamePart('Weihua'))


def test_chunks():
    a1 = Author('Ei-Shazly, M').chunks
    a2 = Author('EI-Shazly, M.').chunks
    assert a1 == a2

    n1 = NamePart('H.H.')
    n2 = NamePart('Hsiao-Hwa')
    n3 = NamePart('Hui')
    assert n1 == n2
    assert n1 != n3
    assert n2 != n3


def _test_match_score():
    corr, expected, cand = corresponding_authors[8]
    corr = Author(corr)
    print
    print corr, corr.chunks
    print
    import difflib
    scored_can = []
    for a in cand:
        p1 = ' '.join([c.name for c in sorted(corr.chunks)])
        p2 = ' '.join([c.name for c in sorted(Author(a).chunks)])
        l = levenshtein(p1, p2)
        l = float(l ** 2) / len(a) / len(corr.name)
        c = corr.match_score(Author(a))
        scored_can.append((
            Author(a),
            c,
            l,
            c - 10 * l,
            corr.distance(Author(a))
        ))

    for a, c, l, c2, d in scored_can:
        print a.chunks, c, l, c2, d

    print
    print 'Score:', max(scored_can, key=lambda c: c[1])
    print 'Leven:', min(scored_can, key=lambda c: c[2])
    print 'Mixed:', max(scored_can, key=lambda c: c[3])
    print 'Implemented:', min(scored_can, key=lambda c: c[4])
    print 'Diff:', difflib.get_close_matches(corr.name, cand, 1)



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Unicode, ForeignKey, Text, Boolean
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.sql.expression import ClauseElement


Base = declarative_base()


def get_or_create(session, model, defaults={}, **kwargs):
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance, False
    else:
        params = {k: v for k, v in kwargs.iteritems()
                  if not isinstance(v, ClauseElement)}
        params.update(defaults)
        instance = model(**params)
        session.add(instance)
        return instance, True


class Institution(Base):
    __tablename__ = 'institution'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(1024), nullable=False)
    country = Column(Unicode(100))

    def get_authors(self, session):
        return (session.query(Person)
                .join(AffiliatedAuthor)
                .filter(AffiliatedAuthor.institution == self))

    def get_publications(self, session):
        return (session.query(Publication)
                .join(AffiliatedAuthor)
                .filter(AffiliatedAuthor.institution == self))

    @property
    def authors(self):
        return self.get_authors(object_session(self))

    @property
    def publications(self):
        return self.get_publications(object_session(self))


class Person(Base):
    __tablename__ = 'person'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(60), nullable=False)

    def get_institutions(self, session):
        return (session.query(Institution)
                .join(AffiliatedAuthor)
                .filter(AffiliatedAuthor.author == self))

    def get_publications(self, session):
        return (session.query(Publication)
                .join(AffiliatedAuthor)
                .filter(AffiliatedAuthor.author == self))

    @property
    def institutions(self):
        return self.get_institutions(object_session(self))

    @property
    def publications(self):
        return self.get_authors(object_session(self))


class Publication(Base):
    __tablename__ = 'publication'

    # Basic attributes
    id = Column(Integer, primary_key=True)
    type = Column(String(32), nullable=True)
    title = Column(Unicode(1024), nullable=False)
    year = Column(Integer)
    times_cited = Column(Integer, nullable=True)

    # Relationships
    def get_institutions(self, session):
        return (session.query(Institution)
                .join(AffiliatedAuthor)
                .filter(AffiliatedAuthor.publication == self))

    @property
    def institutions(self):
        return self.get_institutions(object_session(self))

    def get_authors(self, session):
        return (session.query(Person)
                .join(AffiliatedAuthor)
                .filter(AffiliatedAuthor.publication == self))

    @property
    def authors(self):
        return self.get_authors(object_session(self))

    # Bookkeping metadata
    unique_source_id = Column(String(64), nullable=False, unique=True)
    unparsed_record_format = Column(String(20))
    unparsed_record_value = Column(Text())


class AffiliatedAuthor(Base):
    __tablename__ = 'affiliated_author'

    id = Column(Integer, primary_key=True)
    order = Column(Integer, nullable=False)

    unparsed_institution_name = Column(Unicode(1024))
    institution_id = Column(Integer, ForeignKey('institution.id'))
    institution = relationship(Institution, backref='affiliations')

    unparsed_person_name = Column(Unicode(60), nullable=False)
    person_id = Column(Integer, ForeignKey('person.id'), nullable=False)
    author = relationship(Person, backref='affiliations')

    publication_id = Column(Integer, ForeignKey('publication.id'),
                            nullable=False)
    publication = relationship(Publication, backref='affiliations')

    is_corresponding = Column(Boolean, default=False)



import argparse
from sqlalchemy import create_engine
from irco import models
from irco.logging import sentry


def main():
    argparser = argparse.ArgumentParser('irco-init')
    argparser.add_argument('-v', '--verbose', action='store_true')
    argparser.add_argument('database')

    args = argparser.parse_args()

    sentry.context.merge({
        'tags': {'command': 'irco-init'},
        'extra': {'parsed_arguments': args.__dict__}
    })

    engine = create_engine(args.database, echo=args.verbose)
    models.Base.metadata.create_all(engine)



import re
import logging
from collections import Counter


log = logging.get_logger()


class NamePart(object):
    def __init__(self, name):
        self.name = name
        self._preprocess_abbr()

    def _preprocess_abbr(self):
        if self.name.endswith('.'):
            self.is_abbreviated = True
        elif self.name.isupper():
            self.is_abbreviated = True
            self.name = ''.join((c + '.' for c in self.name))
        else:
            self.is_abbreviated = False

    def __len__(self):
        if self.is_abbreviated:
            return sum((c.isupper() for c in self.name))
        else:
            return len(self.name)

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return 'NamePart({!r})'.format(self.name.encode('utf-8'))

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if not isinstance(other, NamePart):
            return NotImplemented

        if self.isabbrof(other) or other.isabbrof(self):
            return True

        return False

    def __gt__(self, other):
        if not isinstance(other, NamePart):
            return NotImplemented
        return self.name > other.name

    def __lt__(self, other):
        if not isinstance(other, NamePart):
            return NotImplemented
        return self.name < other.name

    def __ge__(self, other):
        if not isinstance(other, NamePart):
            return NotImplemented
        return self.name >= other.name

    def __le__(self, other):
        if not isinstance(other, NamePart):
            return NotImplemented
        return self.name <= other.name

    def isabbrof(self, other):
        pattern = re.escape(self.name).replace('\.', '.+')
        return re.match(pattern, other.name, re.I) is not None


class Author(unicode):
    def __init__(self, name):
        match = re.match(r'^([^\(]+)(?:\(([^@]+@[^ ]+)\))?$', name)
        if not match:
            match = re.match(r'^([^\(]+)\(', name)
            self.name = match.group(0)
            self.email = None
        else:
            self.name, self.email = match.groups()
        self.chunks = self.split_name(self.name)

    @classmethod
    def from_chunks(cls, name, email, chunks):
        self = cls(name)
        self.name = name
        self.email = email
        self.chunks = chunks
        return self

    def split_name(self, name):
        chunks = []
        for c in name.split(u', '):
            chunks += self.split_chunk(c)
        chunks = sorted(chunks, key=lambda p: p.name)
        return tuple(chunks)

    def split_chunk(self, chunk):
        parts = [c.strip() for c in re.split(r'([. ]{1,2}(?!-))', chunk)]

        i = 0

        while i < len(parts):
            if not parts[i]:
                del parts[i]
            elif parts[i] == '.':
                parts[i-1] += parts[i]
                del parts[i]
            else:
                i += 1

        parts = [NamePart(p) for p in parts]
        parts = sorted(parts, key=lambda p: p.name)
        return tuple(parts)

    def __unicode__(self):
        return self.name

    def __repr__(self):
        return 'Author({})'.format(', '.join([repr(c) for c in self.chunks]))

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        if not isinstance(other, Author):
            return False

        if len(self.chunks) != len(other.chunks):
            return False

        self_abbr = Counter([c for c in self.chunks if c.is_abbreviated])
        self_not_abbr = Counter(self.chunks) - self_abbr

        other = Counter(other.chunks)

        r1 = self._find_and_remove(self_not_abbr, other)
        r2 = self._find_and_remove(self_abbr, other)

        return r1 and r2

    def __hash__(self):
        initials = ''.join(n.name[0] for n in self.chunks).lower()
        return hash(initials)

    def _find_and_remove(self, set1, set2):
        while set1:
            n1, c = set1.popitem()
            if c > 1:
                set1[n1] = c - 1
            for n2 in set2:
                if n1.is_abbreviated ^ n2.is_abbreviated:
                    continue
                if n1 == n2:
                    set2.subtract([n2])
                    if not set2[n2]:
                        del set2[n2]
                    break
            else:
                for n2 in set2:
                    if n1 == n2:
                        set2.subtract([n2])
                        if not set2[n2]:
                            del set2[n2]
                        break
                else:
                    return False
        return True

    def find_best_match(self, authors):
        authors = set(authors)

        for a in authors:
            if a == self:
                return a

        candidates = [(a, self.match_score(a) - 10 * self.distance(a))
                      for a in authors]
        match = max(candidates, key=lambda a: a[1])[0]
        log.warning('Fuzzy match search for "{}" matched "{}"'.format(
            self.name, match.name))
        return match

    def match_score(self, other):
        score = 0

        abbr = Counter([c for c in self.chunks if c.is_abbreviated])
        full = Counter(self.chunks) - abbr

        other_abbr = Counter([c for c in other.chunks if c.is_abbreviated])
        other_full = Counter(other.chunks) - other_abbr

        def substract(d1, d2):
            common_sum = 0
            for k1, count1 in d1.items():
                for k2, count2 in d2.items():
                    if k1 == k2:
                        in_common = min(count1, count2)
                        common_sum += in_common
                        count1 -= in_common
                        d2[k2] -= in_common
                        if not count1:
                            break
                d1[k1] = count1
            return common_sum

        score += 10 * substract(full, other_full)
        score += 10 * substract(abbr, other_abbr)
        score += 5 * substract(full, other_abbr)
        score += 5 * substract(abbr, other_full)

        for d in (full, abbr, other_full, other_abbr):
            for k in d.values():
                assert k >= 0
                score -= k * 5

        return score

    def distance(self, other):
        p1 = ' '.join([c.name for c in sorted(self.chunks)])
        p2 = ' '.join([c.name for c in sorted(other.chunks)])
        l = len(self.name) * len(other.name)
        return float(levenshtein(p1, p2) ** 2) / l


def levenshtein(s1, s2):
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)

    previous_row = xrange(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]



from __future__ import print_function
import argparse
import os
import requests
from irco.logging import sentry, get_logger


DOWNLOAD_URL = 'http://apps.webofknowledge.com/OutboundService.do?action=go'
MAX_RECORDS = 500


# http://wokinfo.com  -- Subscriber login >
# http://sub3.webofknowledge.com/error/Error?PathInfo=%2F&Alias=WOK5&Domain=.webofknowledge.com&Src=IP&RouterURL=http%3A%2F%2Fwww.webofknowledge.com%2F&Error=IPError
# requests.get('http://apps.webofknowledge.com/UA_GeneralSearch_input.do?product=UA&search_mode=GeneralSearch&SID=Z18u1itzh2szEsmsje8&preferencesSaved=')


def download(search_id, start, stop, output_stream):
    r = requests.post(DOWNLOAD_URL, stream=True, data={
        'displayCitedRefs': 'true',
        'displayTimesCited': 'true',
        'product': 'WOS',
        'colName': 'WOS',
        'mode': 'OpenOutputService',
        'qid': '2',
        'SID': search_id,
        'format': 'saveToFile',
        'filters': ' '.join([
            # 'USAGEIND',
            # 'AUTHORSIDENTIFIERS',
            'ACCESSION_NUM',
            # 'FUNDING',
            # 'SUBJECT_CATEGORY',
            # 'JCR_CATEGORY',
            # 'LANG',
            # 'IDS',
            # 'PAGEC',
            # 'SABBR',
            # 'CITREFC',
            # 'ISSN',
            # 'PUBINFO',
            # 'KEYWORDS',
            'CITTIMES',
            'ADDRS',
            # 'CONFERENCE_SPONSORS',
            'DOCTYPE',
            # 'CONFERENCE_INFO',
            'SOURCE',
            'TITLE',
            'AUTHORS',
        ]),
        'mark_from': start,
        'mark_to': stop,
        'save_options': 'tabMacUTF8',

        # Unused fields
        # 'selectedIds': '',
        # 'viewType': 'summary',
        # 'mark_id': 'WOS',
        # 'search_mode': 'GeneralSearch',
        # 'locale': 'en_US',
        # 'view_name': 'WOS-summary',
        # 'sortBy': 'PY.D;LD.D;SO.A;VL.D;PG.A;AU.A',
        # 'count_new_items_marked': '0',
        # 'value(record_select_type)': 'range',
        # 'markFrom': '1',
        # 'markTo': '10',
    })

    gen = r.iter_content(1024)

    for chunk in gen:
        output_stream.write(chunk)


def main():
    log = get_logger()

    argparser = argparse.ArgumentParser('irco-scrape')
    argparser.add_argument('search_id')
    argparser.add_argument('output')
    argparser.add_argument('count', type=int)
    args = argparser.parse_args()

    sentry.context.merge({
        'tags': {'command': 'irco-init'},
        'extra': {'parsed_arguments': args.__dict__}
    })

    log.info('arguments_parsed', args=args)

    if not os.path.exists(args.output):
        os.makedirs(args.output)

    digits = len(str(args.count))

    for i, start in enumerate(range(0, args.count, MAX_RECORDS)):
        dest = os.path.join(args.output, 'savedrecs-{:05d}.csv'.format(i))
        end = min(args.count, start + MAX_RECORDS)
        print('{:{}d} - {:{}d} => {}'.format(
            start + 1, digits, end, digits, dest))
        with open(dest, 'wb') as fh:
            download(args.search_id, start + 1, end, fh)



import sys
import itertools
import collections

import networkx as nx


from irco import logging

log = logging.get_logger()


def get_countries(publication):
    publication_countries = set()

    for affiliation in publication.affiliations:
        country = affiliation.institution.country
        if country is None:
            print >>sys.stderr, 'Undefined country for "{}"'.format(
                affiliation.institution.name)
        else:
            publication_countries.add(country)

    return publication_countries


def create(session, publications):
    g = nx.Graph()

    papers_count = collections.Counter()
    collaborations_count = collections.Counter()

    for publication in publications:
        publication_countries = get_countries(publication)
        g.add_nodes_from(publication_countries)
        papers_count.update(publication_countries)

        collaborations = list(itertools.combinations(
            sorted(publication_countries), 2))
        collaborations_count.update(collaborations)
        g.add_edges_from(collaborations)

    # Set papers count
    for country, count in papers_count.iteritems():
        g.node[country]['papers'] = count

    # Set edge weight
    for (c1, c2), count in collaborations_count.iteritems():
        g[c1][c2]['weight'] = count

    return g



import math

from jinja2 import Markup


class Page(object):
    def __init__(self, paginator, num, label=None):
        self.num = num
        self.paginator = paginator
        self.label = label if label else num + 1

    @property
    def is_current(self):
        return self.num == self.paginator.current_page

    def __html__(self):
        if 0 <= self.num < self.paginator.pages:
            link = '<a href="?page={}">{}</a>'.format(self.num + 1, self.label)
            class_ = 'active' if self.is_current else ''
        else:
            link = '<span>{}</span>'.format(self.label)
            class_ = 'disabled'

        return '<li class="{}">{}</li>'.format(class_, link)


class Paginator(object):
    block_size = 7
    limit_size = 2

    def __init__(self, count, per_page, current_page=0):
        self.count = count
        self.per_page = per_page
        self.current_page = current_page
        self.pages = int(math.ceil(1. * count / per_page))

    def next(self):
        return Page(self, self.current_page + 1, 'Next &rarr;')

    def prev(self):
        return Page(self, self.current_page - 1, '&larr; Prev')

    def ellipsis(self):
        return Markup(u'<li class="ellipsis disabled"><span>&hellip;</span></li>')

    def __iter__(self):
        side_size = int(math.ceil(self.block_size / 2))
        start = self.current_page - side_size
        end = self.current_page + side_size

        if start < 0:
            end -= start
        elif end > self.pages:
            start -= end - self.pages

        if start > self.limit_size + 1:
            yield Page(self, 0)
            yield Page(self, 1)
            yield self.ellipsis()
        else:
            start = 0

        for i in range(max(0, start), min(self.pages, end + 1)):
            yield Page(self, i)

        if end < self.pages - self.limit_size - 2:
            yield self.ellipsis()
            yield Page(self, self.pages - 2)
            yield Page(self, self.pages - 1)
        else:
            for i in range(end + 1, self.pages):
                yield Page(self, i)



from __future__ import print_function

import csv
import re
import pprint

from ftfy import fix_text

from irco.authors import Author
from . import base


class LineGetter(object):
    def __init__(self, stream):
        self.value = ''
        self.stream = stream

    def __iter__(self):
        for l in self.stream:
            self.value = l
            l = l.encode('utf-8').rstrip('\r')
            if l.endswith('\t'):
                l = l[:-1]
            yield l

    def get(self):
        return self.value.strip()


class Tokenizer(base.Tokenizer):
    FORMAT = 'webofscience-tsv'
    FIELDS = {
        'TI': 'title',
        'PY': 'year',
        'PT': 'type',
        'C1': 'authors_with_affiliations',
        'TC': 'times_cited'
    }

    def __init__(self, encoding='utf8'):
        self.encoding = encoding

    def tokenize(self, stream):
        self.stream = stream
        self.line = 2
        self.header_row = next(self.stream).rstrip('\r') + '\n'
        self.stream.seek(0)
        self.lines = LineGetter(self.stream)
        self.reader = csv.DictReader(self.lines, delimiter='\t')
        self.reader.fieldnames = [self._get_key(f) for f in
                                  self.reader.fieldnames]
        return self

    def __iter__(self):
        return self

    def next(self):
        row = next(self.reader)
        record = base.Record(
            self.FORMAT,
            self.header_row + self.lines.get()
        )
        record.source = (self.stream.name, self.line)
        record.update(row)
        self.line += 1

        return record

    def _get_key(self, k):
        k = fix_text(unicode(k, self.encoding))
        k = str(k.encode('ascii', 'ignore'))
        k = self.FIELDS.get(k, k)
        return k


class Parser(base.Parser):
    def __init__(self, encoding='utf8'):
        self.encoding = encoding

    def parse_record(self, record):
        for k, v in record.iteritems():
            if k is None:
                raise ValueError('Incorrectly parsed row')
            record[k] = self._get_value(v)
        return record

    def _get_value(self, v):
        v = unicode(v, self.encoding)
        return fix_text(v)


class BaseValuesProcessor(base.Processor):
    pubtypes = {
        'J': 'journal',
        'C': 'conference',
        'B': 'book',
        'P': 'patent',
        'S': 'book in series',
    }

    def process_record(self, record):
        try:
            record['type'] = self.pubtypes[record['type']]
        except KeyError:
            raise ValueError('Unknown publication type: {}'.format(
                record['type']))
        record['times_cited'] = int(record['times_cited'])
        unique_id = record.pop('UT')
        record.unique_source_id = 'wos/' + unique_id.split(':', 1)[1]
        return record


class AffiliationsProcessor(base.Processor):
    splitter = re.compile(r'\[([^\]]+)] ([^;]+)(?:; |$)')

    def initialize(self, pipeline):
        self.pipeline = pipeline
        self.unmatched_authors = open('unmatched_authors.log', 'w')
        pipeline.add_metric('corresponding_author_unmatched',
                            'Records wixth unmatched corr. auth.')
        pipeline.add_metric('corresponding_author_undefined',
                            'Records with undefined corr. auth.')
        pipeline.add_metric('ambiguous_author_affiliations',
                            'Records with ambiguous auth. aff.')

    def process_record(self, record):
        record['institutions'] = []
        record['authors'] = []

        affiliations = self.splitter.findall(
            record['authors_with_affiliations'].strip())

        if not affiliations:
            aut = record['AF'].split('; ')
            aff = record['authors_with_affiliations'].split('; ')
            if len(aut) == len(aff):
                affiliations = list(zip(aut, aff))
            elif len(aff) == 1:
                affiliations = [(a, aff[0]) for a in aut]
            else:
                self.pipeline.inc_metric('ambiguous_author_affiliations')
                print('-' * 80)
                print(u'Ambiguous author affiliations for "{title}":'.format(
                    **record))
                print(' Authors:')
                for a in aut:
                    print('  * {}'.format(a))
                print(' Affiliations:')
                for a in aff:
                    print('  * {}'.format(a))
                print('-' * 80)
                return None

        # TODO: Some authors could have two affiliations! This should be
        # checked here and a warning raised.

        for i, (authors, institution) in enumerate(affiliations):
            record['institutions'].append(institution)

            for a in authors.split('; '):
                author = Author(a)
                record['authors'].append((author, i))

        if record['RP']:
            t = ' (reprint author)'
            corresponding = record['RP'][:record['RP'].find(t)]
            corresponding = Author(corresponding.strip())
            t += ', '
            institution = record['RP'][record['RP'].find(t) + len(t):]

            match = corresponding.find_best_match(
                [a[0] for a in record['authors']])
            if not match:
                self.pipeline.inc_metric('corresponding_author_unmatched')
                print('-' * 80)
                print('No corresponding author match found for:')
                print('  {!r}/{!r}'.format(record['title'],
                                           corresponding.name))
                names = (a[0].name for a in record['authors'])
                pprint.pprint((corresponding.name, 0, tuple(names)),
                              self.unmatched_authors)
                print('-' * 80)
                return None
            else:
                for i, (a, institution_id) in enumerate(record['authors']):
                    if a is match:
                        record['corresponding_author'] = i
                        curr_inst = record['institutions'][institution_id]
                        if institution != curr_inst:
                            record['institutions'].append(institution)
                            record['authors'][i] = (
                                a, len(record['institutions']) - 1)
                        break
        else:
            self.pipeline.inc_metric('corresponding_author_undefined')
            record['corresponding_author'] = 0
            print (u'Undefined corresponding author for "{}", selecting "{}"'
                   .format(record['title'], record['authors'][0][0].name))
        return record


def pipeline(encoding):
    return base.Pipeline(
        Tokenizer(encoding),
        Parser(encoding),
        [
            BaseValuesProcessor(),
            AffiliationsProcessor(),
        ],
        encoding,
    )
